Object subclass: #EnchantingCodeExporter
	instanceVariableNames: 'dict functions indentationLevel sprite codeStream argDepth '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting'!
!EnchantingCodeExporter commentStamp: 'CWB 8/4/2010 22:17' prior: 0!
The EnchantingCodeExported is used to take the scripts of a Scratch sprite and export them into computer code.

dict is a dictionary to convert from the English representation of a code tile to whichever language interests us.

functions is a list of function bodies (in textual format) to be exported.

indentationLevel is how many tabs to put in front of a line of code.

sprite points back to the sprite whose code tiles we are exporting.

codeStream is a stream for placing the code sequences we are building on.

When argDepth is greater than zero, we are within an argument block.  (No tabbing should be performed).



!


!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 8/5/2010 21:29'!
codeSpecFrom: english
	"Returns the specification for the block in the computer language we are exporting to"

	^ dict at: english ifAbsent: [ '>>?', english, '?<<' ].	! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 7/27/2010 19:36'!
englishSpecForSelector: aSelector
	"Returns the specification for the block in the native english"

	^ self sprite blockSpecForSelector: aSelector.	! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 8/5/2010 22:36'!
exportGreenFlagScript: aBlock
	"Emits code for a script triggered by startup"

	self startScript: 'WhenGreenFlagClicked'.
	aBlock nextBlock ifNotNil: [aBlock nextBlock exportCodeWith: self].
	self endScript.! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 7/10/2010 07:37'!
importTranslation: filename
	"Reads a translation dictionary from the specified file"
 
	dict _ ScratchTranslator importTranslation: filename
	! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 8/4/2010 22:18'!
initialize

	super initialize.
	functions _ OrderedCollection new.
	indentationLevel _ 0.
	argDepth _ 0.
	sprite _ nil.! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 8/14/2010 21:58'!
outputTranslationOfTerm: englishTerm
	"Takes an english term (such as appears in a choice argument morph), and translates it onto the code stream"

	codeStream nextPutAll: (self codeSpecFrom: englishTerm).! !

!EnchantingCodeExporter methodsFor: 'as yet unclassified' stamp: 'CWB 8/4/2010 22:25'!
printOn: aStream

	functions do: [:f | aStream nextPutAll: f; cr. ]


! !


!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:54'!
codeStream
	"Gets the stream that code is being stored on"

	^ codeStream! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:55'!
codeStream: aStream
"Sets the stream to store code on.  If nil, creates a new code stream"

	aStream ifNil:
		[ codeStream _ WriteStream on: (String new: 5000) ]
	ifNotNil:
		[ codeStream _ aStream ]! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/21/2010 13:28'!
sprite
	"Grants access to the sprite being exported"

	^ sprite
! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/21/2010 13:28'!
sprite: aSprite
	"Sets the sprite to be operated on."

	sprite _ aSprite
! !


!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 7/28/2010 11:35'!
endCompoundStatement
	"Send this to end a new block of code that was started with startCompoundStatement."

	indentationLevel _ indentationLevel - 1.
	self indent.
	codeStream nextPut: $}; cr.
	
	  ! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/5/2010 22:32'!
endScript
	"Ends the emitting of a script started by sending startScript"

	self endCompoundStatement.
		! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/14/2010 22:28'!
exportArgument: aMorph
	"Takes the given argument to another block, translates if necessary, and outputs the result"

	argDepth _ argDepth + 1.

	"A number, a string, a dropdown -- anything except a code block will be an ArgMorph"
	(aMorph isKindOf: ArgMorph) ifTrue: [
		"codeStream nextPutAll: '[ArgMorph] '. " 
		aMorph exportArgWith: self.
	].

	"Any place a code tile has been inserted is a BlockMorph"
	(aMorph isKindOf: BlockMorph) ifTrue: [
		"codeStream nextPutAll: '[BlockMorph] '.  "
		"codeStream nextPut: $(."
		self exportBlock: aMorph.		"May need to flag that no indent applies"
		codeStream skip: -1.  "remove carriage return"
		"codeStream nextPut: $)."
	].

	argDepth _ argDepth - 1.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/14/2010 22:28'!
exportBlock: aBlock
	"Takes the given block, translates it to code, and outputs the result"

	| englishSpec codeSpec argMapping argIndex |

	aBlock isHatBlock 
		ifTrue: [ englishSpec _ aBlock englishCodeSpec ]
		ifFalse: [ englishSpec _ self englishSpecForSelector: aBlock selector. ].

	englishSpec ifNil: [ codeStream nextPutAll: '->NO ENGLISH FOR ' , aBlock class asString asUppercase , '<-'. ^ nil ].

	codeSpec _ self codeSpecFrom: englishSpec.
	argMapping _ aBlock class argPermutationForSpec: englishSpec withTranslation: codeSpec.

	self indent.

	argIndex _ 1.
	(CommandBlockMorph parseCommandSpec: codeSpec removeSpaces: false ) do: 
		[ :token |
			(CommandBlockMorph isArgSpec: token) 
				ifTrue: [ 
					| arg |
					"Look up the proper argument"
					arg _ aBlock argumentAt: (argMapping at: argIndex)	.
					self exportArgument: arg.
					argIndex _ argIndex + 1.
				] 
				ifFalse: [ codeStream nextPutAll: token. ].
		].

	codeStream cr.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/6/2010 07:23'!
indent

	"Outputs indentation for the current line of code"

	| indentText |

	indentText _ self codeSpecFrom: 'indent'.

	argDepth == 0 ifTrue: [ indentationLevel timesRepeat: [ codeStream nextPutAll: indentText ] ]! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 7/28/2010 10:56'!
processScript: aScript

	"Takes a script (a hat block that starts a list of blocks) and processes it for output from the exporter"

	"At the moment, just retain what type of block this one is."

	self codeStream: nil.		"Create a new code stream"
	aScript exportCodeWith: self.
	
	functions add: codeStream contents.
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 7/28/2010 11:35'!
startCompoundStatement
	"Send this to start a new block of code (like the insides of a while loop). Send endCompoundStatement when finished."

	self indent.
	codeStream nextPut: ${; cr.
	indentationLevel _ indentationLevel + 1.
	  ! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/5/2010 22:39'!
startScript: baseName
	"Starts emitting a script."

	| functionName preFunctionName postFunctionName |

	functionName _ baseName.	"to do: ensure uniqueness"

	"to do: get these as elements from the dictionary"
	preFunctionName _ 'public void '.
	postFunctionName _ '()'.

	self indent.
	codeStream nextPutAll: (preFunctionName, functionName, postFunctionName); cr.

	self startCompoundStatement.

		! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

EnchantingCodeExporter class
	instanceVariableNames: ''!

!EnchantingCodeExporter class methodsFor: 'instance creation' stamp: 'CWB 7/21/2010 13:29'!
newWithSprite: aSprite andTranslationFile: filename
	"Reads in the given file which has instructions on how to translate"
	| e |

	e _ self new.
	e importTranslation: filename.
	e sprite: aSprite.

	^ e
! !
