BlockMorph subclass: #CommandBlockMorph
	instanceVariableNames: 'commandSpec argPermutation argMorphs titleMorph receiver selector isReporter isTimed wantsName wantsPossession numberArgFlags '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scratch-Blocks'!
!CommandBlockMorph commentStamp: '<historical>' prior: 0!
I represent a command or reporter. I have a label and may also have one or more arguments. The label text, number of arguments, and ordering of arguments and keywords are defined by my commandSpec string. This string consists of a sequence of keywords interspersed with argument specifications that indicate where the arguments appear. Here are some possible argument specifications:

	%b	- boolean
	%c	- color
	%C	- color (chosen with eyedropper, not palette)
	%m	- morph reference
	%n	- number
	%s	- string

Examples:

	%b and %b
	%m's color %C sees %C
	repeat %n times
!


!CommandBlockMorph methodsFor: 'initialization' stamp: 'jm 10/22/2007 09:15'!
initialize

	super initialize.
	commandSpec _ ''.
	argMorphs _ OrderedCollection new.
	receiver _ nil.
	selector _ nil.
	isTimed _ false.
! !


!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 8/6/2008 20:59'!
argMorphs

	^ argMorphs
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'ee 3/2/2009 11:53'!
argPermutation: aSet

	argPermutation _ aSet
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 4/1/2009 10:37'!
args
	"Answer my arguments evaluated and permuted into the order expected by the underlying Squeak method."

	^ self organizeArgs: (argMorphs collect: [:m | m evaluate]) asArray
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 8/24/2003 17:05'!
color: aColor

	super color: aColor.
	self layoutChanged.
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 11/20/2002 09:16'!
commandSpec

	^ commandSpec
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 6/12/2008 13:12'!
commandSpec: aString
	"Set my command spec string. See my class comment."
	"CommandBlockMorph new commandSpec: 'forward %n'"

	commandSpec _ aString.
	numberArgFlags _ nil.
	self addLabel.
	self addCommandIcons.
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 5/7/2009 14:16'!
defaultArgs: defaultValues
	"Set the default values of my arguments from the given list."

	| argM defaultValue |
	1 to: (defaultValues size min: argMorphs size) do: [:i |
		argM _ argMorphs at: (argPermutation at: i).
		defaultValue _ defaultValues at: i.
		(argM  isKindOf: ExpressionArgMorph)
			ifTrue: [argM defaultValueFromSpec: defaultValue]
			ifFalse: [argM defaultValue: defaultValue]].

! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'CWB 7/28/2010 10:53'!
exportCodeWith: anExporter
	"Export the code into a computer language."

	anExporter exportBlock: self.

	self nextBlock ifNotNil: [self nextBlock exportCodeWith: anExporter].

! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 1/21/2007 11:43'!
helpScreenName
	"Answer the name of the help screen for this block, or nil if no help is available."

	selector isInfix ifTrue: [
		'+' = selector ifTrue: [^ 'minus'].
		'-' = selector ifTrue: [^ 'minus'].
		'*' = selector ifTrue: [^ 'minus'].
		'/' = selector ifTrue: [^ 'minus'].

		'<' = selector ifTrue: [^ 'lessThan'].
		'=' = selector ifTrue: [^ 'equals'].
		'>' = selector ifTrue: [^ 'greaterThan'].

		'&' = selector ifTrue: [^ 'and'].
		'|' = selector ifTrue: [^ 'or'].

		'\\' = selector ifTrue: [^ 'mod'].

		^ nil].

	"answer the selector with colons mapped to underscores"
	^ selector asString collect: [:ch | ch = $: ifTrue: [$_] ifFalse: [ch]]
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 4/15/2007 20:01'!
isForever

	^ (selector = #doForever) | (selector = #doForeverIf)
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 4/15/2007 19:59'!
isStop

	^ (selector = #doReturn) | (selector = #stopAll)
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 12/10/2003 20:58'!
isTimed

	^ isTimed
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'DaveF 7/10/2003 17:41'!
isTimed: aBoolean
	"Indicates that this block's evaluate method will be called repeatedly throughout some time interval."

	isTimed _ aBoolean.! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'ee 2/28/2009 17:05'!
numberArgFlags
	"Answer an array of booleans indicating which of my arguments must be numbers. Compute on first use and cache the result."

	| argSpecs argM |
	numberArgFlags ifNotNil: [^ numberArgFlags].

	argSpecs _ (CommandBlockMorph parseCommandSpec: commandSpec) select: [:s | CommandBlockMorph isArgSpec: s].
	argSpecs _ self organizeArgs: argSpecs.
	numberArgFlags _ argSpecs collect: [:spec |
		argM _ self uncoloredArgMorphFor: spec.
		(argM isKindOf: ExpressionArgMorph)
			ifTrue: [
				(argM isKindOf: ExpressionArgMorphWithMenu)
					ifTrue: [
						argM isNumberExpression or:
						[#(listIndexMenu listIndexForDeleteMenu) includes: argM menuSelector]]
					ifFalse: [argM isNumberExpression]]
			ifFalse: [false]].

	^ numberArgFlags
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 6/22/2003 21:50'!
receiver

	^ receiver
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 7/7/2004 00:09'!
receiver: anObject
	"Set my receiver object."

	receiver _ anObject.
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'CWB 7/31/2010 22:12'!
saveReferenceImage
	"Takes a screenshot of the block and saves it to the Scratch folder"

	| english filename |

	english _ ScriptableScratchMorph blockSpecForSelector: selector.
	Transcript show: english; space; show: selector asString; cr.
	english ifNil: [ ^ nil ].	"Just in case the selector has no english text"
	filename _ english , '.gif'.
	self imageForm writeGIFFileNamed: filename.
	! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'CWB 7/31/2010 22:28'!
saveReferenceImageOfTarget: aTarget
	"Takes a screenshot of the block and saves it to the Scratch folder"

	| english filename |

	english _ aTarget blockSpecForSelector: selector.
	selector = #/ ifTrue: [ english _ '%n divided-by %n' ].
	selector = #doIfElse ifTrue: [ english _ 'else' ].
	selector = #turnLeft: ifTrue: [ english _ 'turn (left) %n degrees' ].
	selector = #turnRight: ifTrue: [ english _ 'turn (right) %n degrees' ].

	Transcript show: english; space; show: selector asString; cr.
	english ifNil: [ ^ nil ].	"Just in case the selector has no english text"
	filename _ english , '.gif'.
	self imageForm writeGIFFileNamed: filename.
	! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 7/2/2003 11:47'!
selector

	^ selector
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 8/3/2003 23:52'!
selector: aSymbol

	selector _ aSymbol.
! !

!CommandBlockMorph methodsFor: 'accessing' stamp: 'jm 11/30/2007 12:21'!
selectorAndArg
	"Answer an array containing my selector and argument (if any)."

	| args |
	args _ self args.
	^ Array with: selector with: (args size = 0 ifTrue: [nil] ifFalse: [args first])
! !


!CommandBlockMorph methodsFor: 'evaluation' stamp: 'ee 3/2/2009 11:53'!
argumentAt: index
	"Returns the argument morph at the given index."

	^ argMorphs at: (argPermutation at: index).
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'jm 12/10/2003 18:41'!
argumentCount
	"Returns the number of arguments to this command block."

	^ argMorphs size
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'ee 2/28/2009 17:09'!
coerceArgs: argList
	"Answer an arugments array in which all arguments that should be numbers have been coerced to numbers if necessary."

	| args specialCommands numFlags |
	args _ argList asArray.
	specialCommands _ #(
		append:toList: deleteLine:ofList: getLine:ofList: insert:at:ofList: list:contains: setLine:ofList:to:
		lookLike: showBackground:
		playSound: doPlaySoundAndWait
		setVar:to:).

	(specialCommands includes: selector) ifFalse: [
		"ensure args are numbers where numbers are expected"
		numFlags _ self numberArgFlags.
		1 to: args size do: [:i |
			(numFlags at: i) ifTrue: [args at: i put: (args at: i) asNumberNoError]]].

	^ args
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'jm 7/19/2006 11:06'!
evaluate

	^ self evaluateWithArgs: self args
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'jm 6/25/2009 16:05'!
evaluateInfixWithArgs: rawArgs
	"Evalue an infix operator with the given arguments list. Special case for comparison operatores when both operands are strings. Reverse order of comparison operands in RTL languages."
	"Note: The strings 'true' and 'false' in either English or the current language are equal to the boolean value that they name."

	| args result arg1 arg2 |
	args _ rawArgs.

	#= = selector ifTrue: [
		"special case: check for equality between a boolean and the special strings 'true' or 'false'"
		((args first isKindOf: String) and: [args second isKindOf: Boolean]) ifTrue: [
			(args first = 'true') ifTrue: [^ args second = true].
			(args first = 'false') ifTrue: [^ args second = false].
			(args first = (ScratchTranslator translationFor: 'true')) ifTrue: [^ args second = true].
			(args first = (ScratchTranslator translationFor: 'false')) ifTrue: [^ args second = false]].
		((args first isKindOf: Boolean) and: [args second isKindOf: String]) ifTrue: [
			(args second = 'true') ifTrue: [^ args first = true].
			(args second = 'false') ifTrue: [^ args first = false].
			(args second = (ScratchTranslator translationFor: 'true')) ifTrue: [^ args first = true].
			(args second = (ScratchTranslator translationFor: 'false')) ifTrue: [^ args first = false]]].

	(#(< = >) includes: selector)
		ifTrue: [((args first isKindOf: String) or: [args second isKindOf: String]) ifTrue: [
				arg1 _ receiver interpretStringAsNumberIfPossible: args first.
				arg2 _ receiver interpretStringAsNumberIfPossible: args second.
				((arg1 isKindOf: String) and: [arg2 isKindOf: String]) ifTrue: [
					result _ args first compare: args second.
					#< = selector ifTrue: [^ result = 1].
					#= = selector ifTrue: [^ result = 2].
					#> = selector ifTrue: [^ result = 3]].
				((arg1 isKindOf: Number) and: [arg2 isKindOf: Number]) ifTrue: [
					^ arg1 perform: selector with: arg2].
				^ false]]
		ifFalse: [(ScratchTranslator isRTLMath and: [((selector = #\\) not)]) ifTrue: [args _ args reversed]]. "RTLMath operators are RTL"

	(#(& |) includes: selector) ifTrue: [^ args first perform: selector with: args second].

	^ args first asNumberNoError perform: selector with: args second asNumberNoError
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'jm 6/3/2008 17:24'!
evaluateWithArgs: rawArgs
	"Evalue this block with the given argument list."

	| args |
	"special case for math and boolean infix operators"
	selector isInfix ifTrue: [^ self evaluateInfixWithArgs: rawArgs].

	args _ self coerceArgs: rawArgs..

	"special case for unary operators"
	(#(abs not rounded sqrt truncated) includes: selector) ifTrue: [^ args first perform: selector].

	^ receiver perform: selector withArguments: args
! !

!CommandBlockMorph methodsFor: 'evaluation' stamp: 'jm 5/5/2009 17:05'!
organizeArgs: argList
	"Answer an arugments array in which all arguments have been rearranged to the original English permutation"

	| newArgList |
	argPermutation ifNil: [^ argList].
	(#(+ - / * =) includes: selector) ifTrue: [^ argList].
	newArgList _ #().
	1 to: argPermutation size do: [:i |
		newArgList _ newArgList copyWith: (argList at: (argPermutation at: i))].
	^ newArgList
! !


!CommandBlockMorph methodsFor: 'menus' stamp: 'jm 2/27/2009 11:51'!
rightButtonMenu

	| menu sFrame choice spec |
	menu _ CustomMenu new.
	menu add: 'help' action: #presentHelpScreen.

	(owner isKindOf: ScratchBlockPaletteMorph) ifFalse: [
		menu addLine.
		(#(+ - * / \\) includes: selector) ifTrue: [
			#(+ - * / mod) with: #(+ - * / \\) do: [:s :op | menu add: s action: op]].
		(#(< = >) includes: selector) ifTrue: [
			#(< = >) do: [:op | menu add: op action: op]].
		(#(& |) includes: selector) ifTrue: [
			#(and or) with: #(& |) do: [:s :op | menu add: s action: op]].

		menu addLine.
		menu add: 'duplicate' action: #duplicate.
		(self owner isKindOf: BlockMorph) ifFalse: [  "can't yet delete a blocks inside a script"
			menu add: 'delete' action: #delete]].

	sFrame _ self ownerThatIsA: ScratchFrameMorph.
	(sFrame notNil and: [#(sensor: sensorPressed:) includes: selector]) ifTrue: [
		menu addLine.
		menu add: 'show ScratchBoard watcher' action: #showSensorBoard.
		sFrame workPane scratchServer
			ifNil: [menu add: 'enable remote sensor connections' action: #enableRemoteSensors]
			ifNotNil: [menu add: 'disable remote sensor connections' action: #exitScratchSession]].

	DebugMenu ifTrue: [
		menu addLine.
		menu add: 'show tuples' action: #showTuples].
	
	(choice _ menu localize; startUp) ifNil: [^ self].
	(#(presentHelpScreen duplicate delete) includes: choice) ifTrue: [^ self perform: choice].
	choice = #showSensorBoard ifTrue: [sFrame showSensorBoard. ^ self].
	choice = #enableRemoteSensors ifTrue: [sFrame enableRemoteSensors. ^ self].
	choice = #exitScratchSession ifTrue: [sFrame exitScratchSession. ^ self].
	choice = #showTuples ifTrue: [^ self showTuples].

	"change operator"
	spec _ '%n ', choice, ' %n'.
	'\\' = choice	ifTrue: [spec _ ScratchTranslator translationFor: '%n mod %n'].
	'&' = choice	ifTrue: [spec _ ScratchTranslator translationFor: '%b and %b'].
	'|' = choice	ifTrue: [spec _ ScratchTranslator translationFor: '%b or %b'].

	self commandSpec: spec.
	self selector: choice.
! !

!CommandBlockMorph methodsFor: 'menus' stamp: 'jm 2/2/2009 16:51'!
tabToNextField: evt
	"Move the keyboard focus to the next submorph that responds to keyboard events."

	| fields current |
	fields _ argMorphs select: [:m | m isKindOf: ExpressionArgMorph].

	current _ (1 to: fields size) detect: [:i | (fields at: i) labelMorph == evt hand keyboardFocus] ifNone: [fields size].
	evt hand newKeyboardFocus: (fields at: (current \\ fields size) + 1) labelMorph.
! !


!CommandBlockMorph methodsFor: 'processes' stamp: 'jm 12/21/2008 12:25'!
start

	| stage |
	self stop.
	self receiver ifNil: [^ self].
	scratchProc ifNotNil: [^ self].  "this stack is already running"

	(stage _ self receiver ownerThatIsA: ScratchStageMorph) ifNil: [^ self].
	scratchProc _ stage startProcessFor: self.
	self changed.
! !


!CommandBlockMorph methodsFor: 'object i/o' stamp: 'jm 9/24/2003 18:54'!
fieldsVersion

	^ 1
! !

!CommandBlockMorph methodsFor: 'object i/o' stamp: 'jm 9/24/2003 19:01'!
initFieldsFrom: anObjStream version: classVersion

	super initFieldsFrom: anObjStream version: classVersion.
	self initFieldsNamed: #(
		commandSpec
		argMorphs
		titleMorph
		receiver
		selector
		isReporter
		isTimed
		wantsName
		wantsPossession
	) from: anObjStream.
! !

!CommandBlockMorph methodsFor: 'object i/o' stamp: 'jm 9/24/2003 19:01'!
storeFieldsOn: anObjStream

	super storeFieldsOn: anObjStream.
	self storeFieldsNamed: #(
		commandSpec
		argMorphs
		titleMorph
		receiver
		selector
		isReporter
		isTimed
		wantsName
		wantsPossession
	) on: anObjStream.
! !


!CommandBlockMorph methodsFor: 'stack/tuple conversion' stamp: 'ee 3/1/2009 17:42'!
asBlockTuple
	"Answer a tuple (Array) describing this block and its arguments."

	| blockTuple argList arg |
	blockTuple _ Array new: argMorphs size + 1.
	blockTuple at: 1 put: selector.

	argList _ argMorphs.
	argList _ self organizeArgs: argList.
	((#(+ - / * =) includes: selector) and: [ScratchTranslator isRTLMath]) "RTLMath operators are RTL"
		ifTrue: [argList _ argList reversed].
	1 to: argList size do: [:i |
		arg _ argList at: i.
		(arg isKindOf: ArgMorph) | (arg isKindOf: EventTitleMorph)
			ifTrue: [blockTuple at: i + 1 put: arg evaluate]
			ifFalse: [blockTuple at: i + 1 put: arg asBlockTuple]].

	^ blockTuple
! !


!CommandBlockMorph methodsFor: 'private' stamp: 'ee 1/28/2009 14:04'!
addCommandIcons
	"Add additional icons to certain blocks. Do nothing if this isn't one of those blocks."

	| f m |
	#turnLeft: = selector ifTrue: [
		f _ ScratchFrameMorph skinAt: #turnCCW ifAbsent: [^ self].
		m _ self firstSubmorph delete.
		self addMorphFront: (ImageMorph new form: f).
		self addMorphFront: m].

	#turnRight: = selector ifTrue: [
		f _ ScratchFrameMorph skinAt: #turnCW ifAbsent: [^ self].
		m _ self firstSubmorph delete.
		self addMorphFront: (ImageMorph new form: f).
		self addMorphFront: m].

	#stopAll = selector ifTrue: [
		f _ ScratchFrameMorph skinAt: #stopButton ifAbsent: [^ self].
		self addMorphBack: (ImageMorph new form: f)].
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'ee 2/27/2009 12:46'!
addLabel
	"Add a label for my command and arguments. This is the method that really creates the whole block."
	"Details: We remove all the pieces of this block, and then reassemble them, being careful not to destroy any argument morphs that we might still want to use. This is helpful if someone adds/removes parameters from a hat block."

	| tokens args nextArgIndex m next |
	commandSpec ifNil: [^ self].

	"we make sure not to remove the next block."
	next _ self nextBlock.
	self removeAllMorphsIn: (self submorphs reject: [:t | t = next]).

	tokens _ CommandBlockMorph parseCommandSpec: commandSpec.
	args _ tokens select: [:s | CommandBlockMorph isArgSpec: s].

	"remove any excess arg morphs from the list."
	[argMorphs size > args size] whileTrue: [
		m _ argMorphs last.
		argMorphs remove: m.

		"if the argument we're throwing out is anything complicated, we'd like to keep it around in the world."
		(m isKindOf: BlockMorph) ifTrue: [m openInWorld]].

	"create new arg morphs."
	[argMorphs size < args size] whileTrue: [
		argMorphs addLast: (self argMorphFor: (args at: argMorphs size + 1))].

	nextArgIndex _ 1.

	tokens do: [:s |
		(CommandBlockMorph isArgSpec: s)
			ifTrue:	[
				m _ argMorphs at: nextArgIndex.
				nextArgIndex _ nextArgIndex + 1]
			ifFalse:	[
				m _ self labelMorphFor: s].
		self addMorphBack: m].

	self isReporter ifTrue: [
		m _ submorphs first.
		((m isKindOf: ChoiceArgMorph) | (m isKindOf: SpriteArgMorph)) ifTrue: [  "add spacer"
			self addMorphFront: (Morph new color: Color transparent; extent: 2@5)].

		m _ submorphs last.
		((m isKindOf: ChoiceArgMorph) | (m isKindOf: SpriteArgMorph)) ifTrue: [  "add spacer"
			self addMorphBack: (Morph new color: Color transparent; extent: 2@5)]].
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 3/28/2005 17:32'!
argMorphFor: specString
	"Answer an argument morph for the given argument specification string. Set it's color appropriately."

	| m |
	m _ self uncoloredArgMorphFor: specString.
	(m isKindOf: ColorArgMorph)
		ifTrue: [m color: Color random].
	^ m
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 6/2/2009 22:03'!
argMorphToReplace: aMorph
	"Answer a new argument morph to be used to replace the given morph. Answer nil if the given morph is not one of my argMorphs."

	| i argSpecs spec argM defaults v |
	i _ argMorphs indexOf: aMorph ifAbsent: [^ nil].
	argSpecs _ (CommandBlockMorph parseCommandSpec: commandSpec) select: [:s | CommandBlockMorph isArgSpec: s].
	i > argSpecs size ifTrue: [^ nil].
	argM _ self argMorphFor: (argSpecs at: i).

	(#setVar:to: = selector and: [(argSpecs at: i) = '%n']) ifTrue: [
		^ argM stringExpression: '0'].
	spec _ ScriptableScratchMorph blockSpecDict at: selector ifAbsent: [^ argM].

	defaults _ receiver defaultArgsFor: spec.
	i <= defaults size ifTrue: [
		v _ defaults at: (argPermutation indexOf: i).
		(v isKindOf: String)
			ifTrue: [
				(argM isKindOf: ExpressionArgMorph)
					ifTrue: [argM defaultValueFromSpec: v localized]
					ifFalse: [argM defaultValue: v localized]]
			ifFalse: [argM defaultValue: v]].

	^ argM
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 2/26/2009 21:27'!
canBecomeWatcher
	"I determine which blocks can become watchers."

	| i |
	i _ selector asString findAnySubStr: #('mouse' 'key' 'touching' 'distance') startingAt: 1.
	^ (self isReporter) &
	   (self argumentCount <= 1) &
	   ((#(not atRandom abs rounded lineCountOfList: stringLength:) includes: selector) not) &
	   (i > selector asString size)
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 7/4/2008 18:17'!
fixBlockLayout
	"Update the positions of my submorphs."

	| isFirst h x minXForArgs hAdjust nextB w oldPos oldExtent |
	blockLayoutNeeded ifFalse: [^ self].
	super fixBlockLayout.
	oldPos _ self position.
	oldExtent _ self extent.
	hAdjust _ 4.

	isFirst _ true.
	h _ 14.  "minimum height"
	self nonControlFlowSubmorphs do: [:m |
		(m isKindOf: BlockMorph) ifTrue: [m fixBlockLayout].
		(m isKindOf: ArgMorph) ifTrue: [m fixArgLayout].
		(isFirst and: [(m isKindOf: StringMorph) not]) ifTrue: [
			"extra space when an arg overlaps with notch"
			h _ h max: m height + 3.
			hAdjust _ 1].
		isFirst _ false.
		h _ h max: m height].

	((self isKindOf: CBlockMorph) or: [self isKindOf: IfElseBlockMorph])
		ifTrue: [minXForArgs _ self left + PuzzleInset + PuzzleWidth + 8]
		ifFalse: [minXForArgs _ self left].

	h _ h + 10. 
	self isReporter ifTrue: [h _ h - 2. hAdjust _ 2].
	self isStop ifTrue: [h _ h - 5. hAdjust _ 0].

	x _ self left + 5.
	self nonControlFlowSubmorphs do: [:m |
		(m isKindOf: StringMorph)
			ifTrue: [m color: self labelColor]
			ifFalse: [x _ x max: minXForArgs].
		m position: x@(self top + (((h - hAdjust) - m height) // 2)).
		x _ x + m width + 4].

	w _ (x - self left) max: 37.
	((self isKindOf: CBlockMorph) or:
	 [self isKindOf: IfElseBlockMorph])
		ifTrue: [w _ w max: 71].
	self extent: w @ h.
	ScratchTranslator isRTL ifTrue: [
		self left: oldPos x + (oldExtent x - self width)].

	(nextB _ self nextBlock) ifNotNil: [
		(self isKindOf: CBlockMorph) ifFalse: [
			ScratchTranslator isRTL
				ifTrue: [nextB position: (self right - nextB width)@(self bottom - 4)]
				ifFalse: [nextB position: self left@(self bottom - 4)]].
		nextB fixBlockLayout].
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 6/20/2007 19:10'!
labelColor
	"Answer a label color, either black or white, that maximizes the contrast with my color."

	^ LabelColor

"old:
	^ self color luminance > 0.8
		ifTrue: [Color black]
		ifFalse: [Color white]
"
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 6/2/2008 18:27'!
labelMorphFor: aString

	| s |
	s _ aString.
	commandSpec isUnicode ifTrue: [s _ UTF8 withAll: s].
	^ (StringMorph contents: s font: (ScratchFrameMorph getFont: #Label))
		color: self labelColor
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 6/3/2009 16:38'!
mapReceiver: oldObj to: newObj
	"Used in cloning an object to make blocks that refererred to the original object new refer to the new copy."

	receiver == oldObj ifTrue: [receiver _ newObj].
	argMorphs do: [:arg |
		(arg isKindOf: CommandBlockMorph) ifTrue: [
			arg mapReceiver: oldObj to: newObj]].
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 8/31/2006 09:18'!
nextBlock
	"Answer the block following myself in the block sequence or nil if I am the last block in the sequence. Each block is a submorph of the previous block."

	^ self submorphs
		detect: [:m | (m isKindOf: BlockMorph) and: [(argMorphs includes: m) not]]
		ifNone: [nil]
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 3/23/2005 18:02'!
nextBlock: aBlock
	"Add the given block as the block following me. Delete old nextBlock, if any."

	self nextBlock ifNotNil: [self nextBlock delete].
	self addMorph: aBlock.
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 6/3/2008 15:48'!
parseCommandSpec
	"Answer an array of token strings containing my keywords and argument specs."

	| spec result len i j |
	result _ OrderedCollection new.
	spec _ commandSpec.
	(spec isKindOf: UTF8) ifTrue: [spec _ String withAll: spec].
	len _ commandSpec size.

	i _ 1.
	[(i < len) and: [(spec at: i) isSeparator]] whileTrue: [i _ i + 1].
	[i <= len] whileTrue: [
		j _ spec indexOf: $% startingAt: i.
		j > 0
			ifTrue: [
				j > i ifTrue: [result addLast: (spec copyFrom: i to: j - 1)].
				j < len
					ifTrue: [result addLast: (spec copyFrom: j to: j + 1)]
					ifFalse: [result addLast: '%'].
				i _ j + 2]
			ifFalse: [
				result addLast: (spec copyFrom: i to: len).
				i _ len + 1]].

	^ result asArray collect: [:s | s withBlanksTrimmed]
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 12/10/2005 11:04'!
printCodeOn: aStream indent: indent
	"Append a human-readable string for this block on the given stream."

	| nextB hasFinalSpace |
	indent timesRepeat: [aStream nextPutAll: '    '].
	nextB _ self nextBlock.
	hasFinalSpace _ false.
	submorphs do: [:m |
		m ~~ nextB ifTrue: [
			self printCodeSubmorph: m on: aStream.
			hasFinalSpace _ true.
			aStream space]].
	hasFinalSpace ifTrue: [aStream skip: -1].
	aStream cr.
	self nextBlock ifNotNil: [self nextBlock printCodeOn: aStream indent: indent].
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 8/24/2003 17:06'!
replaceArgMorph: oldMorph by: newMorph
	"Replace oldMorph with newMorph in my argMorphs and submorphs lists. This is done when dropping a reporter block onto one of my arguments."

	argMorphs _ argMorphs collect: [:m |
		m == oldMorph ifTrue: [newMorph] ifFalse: [m]].
	self replaceSubmorph: oldMorph by: newMorph.
	self layoutChanged.
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'CWB 7/3/2010 21:23'!
uncoloredArgMorphFor: specString
	"Answer an argument morph for the given argument specification string."

	| code |
	code _ specString at: 2.
	$a = code ifTrue: [^ AttributeArgMorph new choice: 'volume'].
	$A = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #nxtMotorNames; choice: 'left motor'].
	$b = code ifTrue: [^ BooleanArgMorph new].
	$c = code ifTrue: [^ ColorArgMorph new showPalette: true].
	$C = code ifTrue: [^ ColorArgMorph new showPalette: false].  "don't use palette; pick from screen"
	$d = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '0'; menuSelector: #directionMenu].
	$D = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '48'; menuSelector: #midiDrumMenu].
	$e = code ifTrue: [^ EventTitleMorph new].
	$E = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #nxtLightSensorNames; choice: 'light sensor'].
	$f = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #mathFunctionNames; choice: 'sqrt'].
	$g = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #graphicEffectNames; choice: 'color'].
	$H = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #hookupSensorNames].
	$h = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #hookupBooleanSensorNames].
	$I = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '1'; menuSelector: #midiInstrumentMenu].
	$i = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '1'; menuSelector: #listIndexMenu].
	$k = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #keyNames; choice: 'space'].
	$L = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #listVarMenu].
	$l = code ifTrue: [^ ChoiceOrExpressionArgMorph new getOptionsSelector: #costumeNames; choice: 'costume1'].
	$m = code ifTrue: [^ SpriteArgMorph new].
	$M = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #motorNames].
	$n = code ifTrue: [^ ExpressionArgMorph new numExpression: '10'].
	$N = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '60'; menuSelector: #noteSelector].
	$q = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #nxtLightSensorSubMenu; choice: 'brightness'].	
	$s = code ifTrue: [^ ExpressionArgMorph new stringExpression: ''].
	$S = code ifTrue: [^ ChoiceOrExpressionArgMorph new getOptionsSelector: #soundNames; choice: 'pop'].
	$v = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #varNamesMenu; choice: ''].
	$W = code ifTrue: [^ ChoiceArgMorph new getOptionsSelector: #motorDirection].
	$x = code ifTrue: [^ ChoiceOrExpressionArgMorph new getOptionsSelector: #sceneNames; choice: ''].
	$y = code ifTrue: [^ ExpressionArgMorphWithMenu new numExpression: '1'; menuSelector: #listIndexForDeleteMenu].

	^ ExpressionArgMorph new numExpression: '10'
! !

!CommandBlockMorph methodsFor: 'private' stamp: 'jm 3/15/2003 11:23'!
updateReferencesUsing: aDictionary
	"Update my arg morphs."

	super updateReferencesUsing: aDictionary.
	argMorphs _ argMorphs collect: [:m | aDictionary at: m ifAbsent: [m]].
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

CommandBlockMorph class
	instanceVariableNames: ''!

!CommandBlockMorph class methodsFor: 'utilities' stamp: 'ee 3/28/2009 20:07'!
argPermutationForSpec: spec withTranslation: translatedSpec
	"CommandBlockMorph argPermutationForSpec: '%a of %m of %m of %t' withTranslation: '%a de %m de %m de %t'"
	"CommandBlockMorph argPermutationForSpec: '%a of %m of %m of %t' withTranslation: '%t de %m2 de %m1 de %a'"

	| parsedSpec parsedTranslatedSpec parsedSpecArgs parsedTranslatedSpecArgs permutation a copiesSize n originalPermutation |

	parsedSpec _ CommandBlockMorph parseCommandSpec: spec.
	parsedSpecArgs _ parsedSpec select: [:s | CommandBlockMorph isArgSpec: s].
	parsedTranslatedSpec _ CommandBlockMorph parseCommandSpec: translatedSpec.
	parsedTranslatedSpecArgs _ parsedTranslatedSpec select: [:s | CommandBlockMorph isArgSpec: s].

	permutation _ #().
	1 to: parsedSpecArgs size do: [:i | permutation _ permutation copyWith: i].
	originalPermutation _ permutation.

	"no change"
	(parsedSpecArgs = parsedTranslatedSpecArgs) ifTrue: [
			^ originalPermutation].

	"otherwise, calculate permutation"
	1 to: parsedSpecArgs size do: [:i |
		a _ parsedSpecArgs at: i.
		copiesSize _ (parsedSpecArgs select: [:s | (s = a)]) size.
		(copiesSize = 1)
			ifTrue:[permutation at: i put: (parsedTranslatedSpecArgs indexOf: a
				ifAbsent: [Transcript show: 'Permutation failed: ', spec.
						^ originalPermutation])]
			ifFalse:["count copies until now"
				n _ ((parsedSpecArgs copyFrom: 1 to: i) select: [:s | (s = a)]) size.
				permutation at: i put: (parsedTranslatedSpecArgs indexOf: (a,n asString)
					ifAbsent: [Transcript show: 'Permutation failed: ', spec.
						^ originalPermutation])]].
	^ permutation
! !

!CommandBlockMorph class methodsFor: 'utilities' stamp: 'ee 2/28/2009 13:39'!
isArgSpec: aString
	"Answer true if the given string is an argument specification."

	^ ((aString size > 1) and: [aString first = $%])
! !

!CommandBlockMorph class methodsFor: 'utilities' stamp: 'CWB 8/4/2010 22:11'!
parseCommandSpec: aCommandSpec
	"Answer an array of token strings containing my keywords and argument specs."

	^ self parseCommandSpec: aCommandSpec removeSpaces: true.		
! !

!CommandBlockMorph class methodsFor: 'utilities' stamp: 'CWB 8/4/2010 22:10'!
parseCommandSpec: aCommandSpec removeSpaces: noSpaces
	"Answer an array of token strings containing my keywords and argument specs."

	| result len i j spec k |
	result _ OrderedCollection new.
	spec _ aCommandSpec.
	(spec isKindOf: UTF8) ifTrue: [spec _ String withAll: spec].
	len _ aCommandSpec size.

	i _ 1.
	[(i < len) and: [(spec at: i) isSeparator]] whileTrue: [i _ i + 1].
	[i <= len] whileTrue: [
		j _ spec indexOf: $% startingAt: i.
		j > 0
			ifTrue: [
				j > i ifTrue: [result addLast: (spec copyFrom: i to: j - 1)].
				j < len
					ifTrue: [k _ j + 1.
						((spec at: k) isLetter)
							ifTrue: [[(k+1 <= len) and: [(spec at: k+1) isDigit]] whileTrue:[
								k _ k + 1].
								(k >= len)
									ifTrue: [result addLast: (spec copyFrom: j to: len).
										i _ len + 1]
									ifFalse: [result addLast: (spec copyFrom: j to: k).
										i _ k+1]]
							ifFalse: [result addLast: '%'.
										i _ j + 1]]
					ifFalse: [result addLast: '%'.
						i _ j + 2]]
			ifFalse: [
				result addLast: (spec copyFrom: i to: len).
				i _ len + 1]].

	noSpaces ifFalse: [ ^result ].
	result _ result select: [:s | s ~= ' '].
	^ result asArray collect: [:s | s withBlanksTrimmed]
! !
